import asyncio
import datetime
import json
import logging
import os
from concurrent.futures import ThreadPoolExecutor

import requests
from sagemaker_jupyterlab_extension_common.util.environment import (
    Environment,
)

from amazon_sagemaker_jupyter_ai_q_developer.constants import (
    Q_DEVELOPER_SETTINGS_ENDPOINT,
)
from amazon_sagemaker_jupyter_ai_q_developer.exceptions import ServerExtensionException
from amazon_sagemaker_jupyter_ai_q_developer.file_cache_manager import FileCacheManager

logging.basicConfig(format="%(levelname)s: %(message)s")


def get_settings_property_value(settings_data, key, secondary_default):
    default_value = (
        settings_data.get("schema", {})
        .get("properties", {})
        .get(key, {})
        .get("default", secondary_default)
    )
    value = settings_data.get("settings", {}).get(key, default_value)
    return value


def get_q_dev_settings():
    return requests.get(Q_DEVELOPER_SETTINGS_ENDPOINT)


async def get_settings_async(metrics):
    settings = {
        "share_content_with_aws": False,
        "suggestions_with_code_references": False,
        "telemetry_enabled": False,
        "log_level": "ERROR",
    }
    get_settings_start_time = datetime.datetime.now()
    try:
        # Q_DEVELOPER_SETTINGS_ENDPOINT is in parity with base url from SMD so in local development
        # it won't be able to fetch settings and also locally headers have to be manually appended for this to work
        executor = ThreadPoolExecutor(max_workers=1)
        loop = asyncio.get_running_loop()
        response = await loop.run_in_executor(executor, get_q_dev_settings)
        settings_data = response.json()

        settings["share_content_with_aws"] = get_settings_property_value(
            settings_data, "shareCodeWhispererContentWithAWS", False
        )
        settings["suggestions_with_code_references"] = get_settings_property_value(
            settings_data, "suggestionsWithCodeReferences", False
        )
        settings["telemetry_enabled"] = get_settings_property_value(
            settings_data, "codeWhispererTelemetry", False
        )
        settings["log_level"] = get_settings_property_value(
            settings_data, "codeWhispererLogLevel", "ERROR"
        )
    except Exception as e:
        logging.warning(f"Can not read Q Developer settings, using default values. Error: {e}")
    finally:
        elapsed = datetime.datetime.now() - get_settings_start_time
        metrics.put_metric("SettingsLatency", int(elapsed.total_seconds() * 1000), "Milliseconds")
    return settings


def get_q_dev_profile_arn(environment: Environment):
    if environment == Environment.STUDIO_IAM or environment == Environment.MD_IAM:
        return "dummy_profile_arn"  # Not needed as of now for Free tier
    return extractor("~/.aws/amazon_q/q_dev_profile.json", lambda d: d["q_dev_profile_arn"])


def get_md_q_enabled():
    return extractor("~/.aws/amazon_q/settings.json", lambda d: d["q_enabled"])


def extractor(file_path=None, value_extractor=None):
    content = json.loads(FileCacheManager().get_cached_file_content(os.path.expanduser(file_path)))
    val = value_extractor(content)

    if isinstance(val, bool):
        return value_extractor(content)
    if val is None or not val.strip():
        raise ServerExtensionException(f"No value found in {file_path}.")
    return value_extractor(content)


def is_md_environment(environment: Environment):
    return environment in [
        Environment.MD_IAM,
        Environment.MD_IDC,
        Environment.MD_SAML,
    ]


def get_ide_category(environment: Environment):
    if is_md_environment(environment):
        return "JUPYTER_MD"
    else:
        return "JUPYTER_SM"


def getQTelemetryInteractionType(interactionType):
    if interactionType == "copy":
        return "COPY_SNIPPET"
    elif interactionType == "insert-above":
        return "INSERT_AT_CURSOR"
    elif interactionType == "insert-below":
        return "INSERT_AT_CURSOR"
    elif interactionType == "replace":
        return "INSERT_AT_CURSOR"
    elif interactionType == "upvote":
        return "UPVOTE"
    elif interactionType == "downvote":
        return "DOWNVOTE"
    else:
        return "UNKNOWN"
